import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;

/**
 * Utility class for manipulating BufferedImages.
 */
public class BufferedImages {
	private BufferedImages() {}
	
	/**
	 * Returns an array of integer pixels in the default RGB color model
	 * (BufferedImage.TYPE_INT_ARGB) and default sRGB color space that
	 * comprise the entire image.
	 * 
	 * This method is equivalent to calling
	 * <blockquote>
	 *     <tt>BufferedImage.getRGB(0, 0, img.getWidth(), img.getHeight(), rgbArray, 0, img.getWidth())</tt>
	 * </blockquote>
	 * except that it is faster when <tt>img</tt> is already in
	 * the native ARGB color model.
	 * 
	 * The <tt>offset</tt> of the output array is assumed to be <tt>0</tt>, and
	 * <tt>scansize</tt> is assumed to be <tt>img.getWidth()</tt>.
	 * 
	 * With a specific coordinate (x, y) in the image, the ARGB pixel can
	 * be accessed in this way:
	 * <blockquote>
	 *     <tt>pixel = rgbArray[y*img.getWidth() + x]</tt>
	 * </blockquote>
	 * 
	 * @param img		the image whose data is to be extracted
	 * @param rgbArray	if not <tt>null</tt> and is of the correct length,
	 * 					the pixels are written here
	 * @return			array of ARGB pixels
	 * @see {@link java.awt.image.BufferedImage#getRGB(int,int,int,int,int[],int,int)}
	 */
	public static int[] getAllRGB(BufferedImage img, int[] rgbArray) {
		/* Try to use optimized ARGB access strategy if the image is in the correct format */
		if (img.getColorModel().equals(ColorModel.getRGBdefault())) {
			DataBuffer dataBuf = img.getRaster().getDataBuffer();
			int numPixels = dataBuf.getSize();
			
			// Ensure output pixel array is allocated and is of the correct length
			if ((rgbArray == null) || (rgbArray.length != numPixels)) {
				rgbArray = new int[numPixels];
			}
			
			// Copy all packed ARGB pixels (in the sRGB color model) from
			// data buffer to the output pixel array
			for (int i=0; i<numPixels; i++) {
				rgbArray[i] = dataBuf.getElem(i);
			}
			
			return rgbArray;
		}
		
		/* Fallback to using general ARGB access strategy */
		return img.getRGB(0, 0, img.getWidth(), img.getHeight(), rgbArray, 0, img.getWidth());
	}
	
	/**
	 * Sets an array of integer pixels in the default RGB color model
	 * (BufferedImage.TYPE_INT_ARGB) and default sRGB color space to
	 * the set of pixels that comprise the entire image.
	 * 
	 * This method is equivalent to calling
	 * <blockquote>
	 *     <tt>BufferedImage.setRGB(0, 0, img.getWidth(), img.getHeight(), rgbArray, 0, img.getWidth())</tt>
	 * </blockquote>
	 * except that it is faster when <tt>img</tt> is already in
	 * the native ARGB color model.
	 * 
	 * The <tt>offset</tt> of the input array is assumed to be <tt>0</tt>, and
	 * <tt>scansize</tt> is assumed to be <tt>img.getWidth()</tt>.
	 * 
	 * With a specific coordinate (x, y) in the image, the ARGB pixel can
	 * be accessed in this way:
	 * <blockquote>
	 *     <tt>pixel = rgbArray[y*img.getWidth() + x]</tt>
	 * </blockquote>
	 * 
	 * @param img		the image whose data is to be changed
	 * @param rgbArray	the ARGB pixels
	 * @see {@link java.awt.image.BufferedImage#setRGB(int,int,int,int,int[],int,int)}
	 */
	public static void setAllRGB(BufferedImage img, int[] rgbArray) {
		/* Try to use optimized ARGB access strategy if the image is in the correct format */
		if (img.getColorModel().equals(ColorModel.getRGBdefault())) {
			DataBuffer dataBuf = img.getRaster().getDataBuffer();
			int numPixels = dataBuf.getSize();
			
			// Ensure input pixel array is of the expected size
			if (rgbArray.length != numPixels)
				throw new IllegalArgumentException(
					"expected rgbArray to have "+numPixels+" element(s)");
			
			// Copy all packed ARGB pixels (in the sRGB color model) from
			// the output pixel array to data buffer
			for (int i=0; i<numPixels; i++) {
				dataBuf.setElem(i, rgbArray[i]);
			}
			
			return;
		}
		
		/* Fallback to using general ARGB access strategy */
		img.setRGB(0, 0, img.getWidth(), img.getHeight(), rgbArray, 0, img.getWidth());
	}
}