// TODO: modify OtsuThresholdingAlgorithm to use this class
public class Histograms {
	private Histograms() {}
	
	// ### HISTOGRAMS ###
	
	/**
	 * Constructs a histogram of the frequencies of the shades of 
	 * gray present in <tt>inGrayPixels</tt>.
	 * 
	 * Given a specific gray-level (between 0 and 255) that is present
	 * in the input image, the number of input pixels with that level
	 * can be determined from the returned histogram in this way:
	 * <blockquote>
	 *     <tt>levelFreq = outHistogram[level]</tt>
	 * </blockquote>
	 * 
	 * @param inGrayPixels		set of grayscale pixels in an image
	 * 							(each pixel value must be in the range 0-255)
	 * @param outHistogram		if not <tt>null</tt> and is of length <tt>256</tt>,
	 * 							the histogram is stored here
	 * @return					histogram of gray-level frequencies in the
	 * 							input grayscale image; always of length <tt>256</tt>
	 */
	public static int[] calculateHistogramOfGrayLevels(int[] inGrayPixels, int[] outHistogram) {
		// Ensure a buffer is allocated to hold the output histogram
		if ((outHistogram == null) || (outHistogram.length != 256)) {
			outHistogram = new int[256];
		}
		
		// Initialize the output histogram
		for (int i=0; i<256; i++) {
			outHistogram[i] = 0;
		}
		
		// Calculate the output histogram
		try {
			for (int i=0, n=inGrayPixels.length; i<n; i++) {
				outHistogram[inGrayPixels[i]]++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new IllegalArgumentException(
				"inGrayPixels must only contain gray values between 0 and 255");
		}
		
		return outHistogram;
	}
	
	/**
	 * @return		the number of samples in the specified histogram.
	 */
	public static int getNumSamples(int[] histogram) {
		return getNumSamples(histogram, 0, histogram.length);
	}
	
	/**
	 * @return		the number of the samples in the interval [startLevel,limitLevel)
	 * 				of the specified histogram
	 */
	public static int getNumSamples(
		int[] histogram,
		int startLevel, int limitLevel)
	{
		int numSamples = 0;
		for (int i=startLevel; i<limitLevel; i++) {
			numSamples += histogram[i];
		}
		return numSamples;
	}
	
	/**
	 * @return		the mean value of the samples in the specified histogram.
	 */
	public static int getMean(int[] histogram) {
		return getMean(histogram, 0, histogram.length);
	}
	
	/**
	 * @return		the mean value of the samples in the interval [startLevel,limitLevel)
	 * 				of the specified histogram
	 */
	public static int getMean(
		int[] histogram,
		int startLevel, int limitLevel)
	{
		int sumSamples = getSum(histogram, startLevel, limitLevel);
		int numSamples = getNumSamples(histogram, startLevel, limitLevel);
		return getMean(sumSamples, numSamples);
	}
	
	public static int getMean(
		int sumSamples,
		int numSamples)
	{
		return (numSamples != 0)
			? sumSamples/numSamples
			: 0;
	}
	
	/**
	 * @return		the sum of the samples in the specified histogram.
	 */
	public static int getSum(int[] histogram) {
		return getSum(histogram, 0, histogram.length);
	}
	
	/**
	 * @return		the sum of the samples in the interval [startLevel,limitLevel)
	 * 				of the specified histogram
	 */
	public static int getSum(
		int[] histogram,
		int startLevel, int limitLevel)
	{
		int sum = 0;
		for (int i=startLevel; i<limitLevel; i++) {
			int curSample = i;
			int curFreq = histogram[i];
			sum += curSample * curFreq;
		}
		return sum;
	}
	
	public static double getSampleStandardDeviation(int[] histogram) {
		return getSampleStandardDeviation(histogram, 0, histogram.length);
	}
	
	public static double getSampleStandardDeviation(
		int[] histogram,
		int startLevel, int limitLevel)
	{
		return getSampleStandardDeviation(
			histogram,
			getNumSamples(histogram, startLevel, limitLevel),
			startLevel, limitLevel);
	}
	
	public static double getSampleStandardDeviation(
		int[] histogram,
		int numSamples,
		int startLevel, int limitLevel)
	{
		return getSampleStandardDeviation(
			histogram,
			numSamples,
			getMean(histogram, startLevel, limitLevel),
			startLevel, limitLevel);
	}
	
	public static double getSampleStandardDeviation(
		int[] histogram,
		int numSamples,
		int mean,
		int startLevel, int limitLevel)
	{
		return getSampleStandardDeviation(
			numSamples,
			getSumSquaredDeviations(histogram, mean, startLevel, limitLevel));
	}
	
	public static double getSampleStandardDeviation(
		int numSamples,
		int sumSquaredDeviations)
	{
		return Math.sqrt(sumSquaredDeviations/(numSamples-1));
	}
	
	/**
	 * @return		the sum of (X - mean)^2 where X is a sample in the specified histogram
	 */
	public static int getSumSquaredDeviations(
		int[] histogram,
		int mean)
	{
		return getSumSquaredDeviations(histogram, mean, 0, histogram.length);
	}
	
	/**
	 * @return		the sum of (X - mean)^2 where X is a sample in the
	 * 				interval [startLevel,limitLevel) of the specified histogram
	 */
	public static int getSumSquaredDeviations(
		int[] histogram,
		int mean,
		int startLevel, int limitLevel)
	{
		int sumSquareDevs = 0;
		for (int i=startLevel; i<limitLevel; i++) {
			int curSample = i;
			int curFreq = histogram[i];
			sumSquareDevs += square(curSample - mean) * curFreq;
		}
		return sumSquareDevs;
	}
	
	// ### PROBABILITY HISTOGRAMS ###
	
	/**
	 * Constructs a probability histogram of the relative frequencies (%s) of
	 * the data represented by the input histogram.
	 * 
	 * Given a specific sample of the input histogram, the relative frequency (%)
	 * of the sample can be determined from the returned probability histogram
	 * in this way:
	 * <blockquote>
	 *     <tt>sampleRelFreq = outProbabilityHistogram[sampleOffInHistogram]</tt>
	 * </blockquote>
	 * 
	 * @param inHistogram				histogram describing the data to be analyzed
	 * @param numSamplesInHistogram		the total number of samples in the data to be analyzed
	 * @param outProbabilityHistogram	if not <tt>null</tt> and is of length matching
	 * 									that of <tt>inHistogram</tt>, the output
	 * 									probability histogram is stored here
	 * @return							probability histogram of the relative frequencies
	 * 									of each sample in the input histogram;
	 * 									always has a length that matches that of <tt>inHistogram</tt>
	 */
	public static double[] calculateProbabilityHistogramWithHistogram(
		int[] inHistogram, int numSamplesInHistogram, double[] outProbabilityHistogram)
	{
		int numFreqsInHistogram = inHistogram.length;
		
		// Ensure a buffer is allocated to hold the output probability histogram
		if ((outProbabilityHistogram == null) || (outProbabilityHistogram.length != numFreqsInHistogram)) {
			outProbabilityHistogram = new double[numFreqsInHistogram];
		}
		
		// Calculate the output probability histogram
		double numSamplesInHistogramAsDouble = (double) numSamplesInHistogram;
		for (int i=0; i<numFreqsInHistogram; i++) {
			outProbabilityHistogram[i] = ((double) inHistogram[i])/numSamplesInHistogramAsDouble;
		}
		
		return outProbabilityHistogram;
	}
	
	/**
	 * Similar to {@link #calculateProbabilityHistogramWithHistogram(int,int,double[])},
	 * but internally computes the number of samples in the input histogram.
	 */
	public static double[] calculateProbabilityHistogramWithHistogram(
		int[] inHistogram, double[] outProbabilityHistogram)
	{
		return calculateProbabilityHistogramWithHistogram(
			inHistogram,
			getNumSamples(inHistogram),
			outProbabilityHistogram);
	}
	
	// ### MISC ###
	
	private static int square(int n) {
		return n*n;
	}
}