import java.io.File;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Handles the loading of plugins.
 */
public class ImageFilterDemoPlugins {
	private final static String PLUGIN_DIR_NAME = "plugins";
	
	private final static Class[] SUPPORTED_FILTER_TYPES = new Class[] {
		PixelFilter.class,
		PixelBufferFilter.class, PixelBufferFilter2.class,
		PixelBufferToBufferedImageFilter.class,
		BufferedImageFilter.class, BufferedImageFilter2.class,
		ImageFilter.class
	};
	
	// ### PLUGIN LISTS ###
	
	public final static List filterPluginClasses;		// List<Class<SUPPORTED_FILTER_TYPE>>
	public final static List unknownPluginClasses;		// List<Class>
	
	// ### LOADING PLUGINS ###
	
	static {
		/* Search for plugins */
		List pluginClasses;					// List<Class>
		FIND_PLUGIN_CLASSES: {
			final File pluginDir;
			{
				File currentDir;
				try {
					currentDir = new File(System.getProperty("user.dir"));
				} catch (SecurityException e) {
					System.err.println("Unable to load plugins: Location of plugin directory could not be determined");
					e.printStackTrace(System.err);
					
					pluginClasses = Collections.EMPTY_LIST;
					break FIND_PLUGIN_CLASSES;
				}
				
				pluginDir = new File(currentDir, PLUGIN_DIR_NAME);
			}
			
			String[] pluginDirFilenames = pluginDir.list();
			if (pluginDirFilenames == null) {
				// I/O error
				System.err.println("I/O error while reading plugins directory: "+pluginDir);
				pluginDirFilenames = new String[0];
			}
			
			List pluginDirClassnames;		// List<String>
			{
				pluginDirClassnames = new ArrayList();
				for (int i=0, n=pluginDirFilenames.length; i<n; i++) {
					String curFilename = pluginDirFilenames[i];
					if (curFilename.endsWith(".class")) {
						String curClassname = curFilename.substring(0, curFilename.length()-".class".length());
						pluginDirClassnames.add(curClassname);
					}
				}
				Collections.sort(pluginDirClassnames, String.CASE_INSENSITIVE_ORDER);
			}
			
			/** Specialized classloader that loads classes from the plugins directory */
			ClassLoader pluginClassLoader = new ClassLoader() {
				protected Class findClass(String name) throws ClassNotFoundException {
					/* Look for a plugin class with a matching name */
					File classfilePath = new File(pluginDir, name+".class");
					DataInputStream classfileIn;
					try {
						classfileIn = new DataInputStream(new FileInputStream(classfilePath));
					} catch (FileNotFoundException e) {
						ClassNotFoundException e2 = new ClassNotFoundException(name);
						e2.initCause(e);
						throw e2;
					}
					
					byte[] b;
					try {
						b = new byte[(int) classfilePath.length()];
						try {
							classfileIn.readFully(b);
						} catch (IOException e) {
							// XXX: I/O errors are propagated to the caller as a ClassNotFoundException
							ClassNotFoundException e2 = new ClassNotFoundException(name);
							e2.initCause(e);
							throw e2;
						}
					} finally {
						try {
							classfileIn.close();
						} catch (IOException e) {}
					}
					return super.defineClass(name, b, 0, b.length);
				}
			};
			
			pluginClasses = new ArrayList();
			for (int i=0, n=pluginDirClassnames.size(); i<n; i++) {
				String curClassname = (String) pluginDirClassnames.get(i);
				
				Class curClass;
				try {
					curClass = pluginClassLoader.loadClass(curClassname);
				} catch (ExceptionInInitializerError e) {
					System.err.println(
						"Unable to load plugin \""+curClassname+"\": "+
						"Exception was thrown by initializer:");
					e.printStackTrace(System.err);
					continue;
				} catch (LinkageError e) {
					System.err.println(
						"Unable to load plugin \""+curClassname+"\": "+
						"Linkage error:");
					e.printStackTrace(System.err);
					continue;
				} catch (ClassNotFoundException e) {
					System.err.println(
						"Unable to load plugin \""+curClassname+"\": "+
						"Unable to locate class");
					continue;
				} catch (Throwable e) {
					System.err.println(
						"Unable to load plugin \""+curClassname+"\": "+
						"Unknown error:");
					e.printStackTrace(System.err);
					continue;
				}
				
				pluginClasses.add(curClass);
			}
		}
		
		/* Categorize plugins by type */
		{
			filterPluginClasses = new ArrayList();
			unknownPluginClasses = new ArrayList();
		NEXT_PLUGIN_CLASS:
			for (int i=0, n=pluginClasses.size(); i<n; i++) {
				Class curPluginClass = (Class) pluginClasses.get(i);
				
				// Check whether the plugin is a filter-plugin
				for (int j=0; j<SUPPORTED_FILTER_TYPES.length; j++) {
					if (SUPPORTED_FILTER_TYPES[j].isAssignableFrom(curPluginClass)) {
						filterPluginClasses.add(curPluginClass);
						continue NEXT_PLUGIN_CLASS;
					}
				}
				
				// Plugin is of unknown type
				unknownPluginClasses.add(curPluginClass);
				System.err.println("Plugin \""+curPluginClass.getName()+"\" is of unknown type");
			}
		}
	}
	
	// ### INIT ###
	
	private ImageFilterDemoPlugins() {}
}