/**
 * Utility class for manipulating pixel values.
 */
public class Pixels {
	private Pixels() {}
	
	// ### ARGB PIXELS ###
	
	public static int getAlpha(int argbOrAhsbPixel) {
		return ((argbOrAhsbPixel >>> 24) & 0xFF);
	}
	
	public static int getRed(int argbPixel) {
		return ((argbPixel >>> 16) & 0xFF);
	}
	
	public static int getGreen(int argbPixel) {
		return ((argbPixel >>> 8) & 0xFF);
	}
	
	public static int getBlue(int argbPixel) {
		return ((argbPixel >>> 0) & 0xFF);
	}
	
	public static int packPixel(int alpha, int red, int green, int blue) {
		return packARGB(alpha, red, green, blue);
	}
	
	public static int packARGB(int alpha, int red, int green, int blue) {
		return (alpha << 24) | (red << 16) | (green << 8) | (blue << 0);
	}
	
	// ### AHSB PIXELS ###
	
	public static int getHue(int ahsbPixel) {
		return ((ahsbPixel >>> 16) & 0xFF);
	}
	
	public static int getSaturation(int ahsbPixel) {
		return ((ahsbPixel >>> 8) & 0xFF);
	}
	
	public static int getBrightness(int ahsbPixel) {
		return ((ahsbPixel >>> 0) & 0xFF);
	}
	
	public static int packAHSB(int alpha, int hue, int saturation, int brightness) {
		return (alpha << 24) | (hue << 16) | (saturation << 8) | (brightness << 0);
	}
	
	// ### COLOR SPACE CONVERSIONS ###
	
	public static int ARGBToAHSB(int argbPixel) {
		int red = Pixels.getRed(argbPixel);
		int green = Pixels.getGreen(argbPixel);
		int blue = Pixels.getBlue(argbPixel);
		
		int max = Math.max(red, Math.max(green, blue));
		int min = Math.min(red, Math.min(green, blue));
		int range = max - min;
		
		int rawHue = -1;
		if (range != 0) {
			if (max == red)
				rawHue = 60*(green - blue)/(range) + 0;
			if (max == green)
				rawHue = 60*(blue - red)/(range) + 120;
			if (max == blue)
				rawHue = 60*(red - green)/(range) + 240;
		} else {
			// XXX: this is the case where the HUE becomes undefined
			rawHue = 0;
		}
		rawHue %= 360;
		
		int alpha = Pixels.getAlpha(argbPixel);
		int hue = rawHue*255/360;
		int saturation = (max != 0)
			? (range)*255/max
			: 0;	// XXX: this is the case where the SATURATION becomes undefined
		int brightness = max;
		
		return packAHSB(alpha, hue, saturation, brightness);
	}
	
	// ### GRAYSCALE IMAGES ###
	
	/**
	 * Converts the specified ARGB pixel to its level of gray, between 0 and 255.
	 */
	public static int ARGBToGray(int argbPixel) {
		int red = Pixels.getRed(argbPixel);
		int green = Pixels.getGreen(argbPixel);
		int blue = Pixels.getBlue(argbPixel);
		return (red+green+blue)/3;
	}
	
	/**
	 * Converts the specified ARGB pixels to grayscale pixels.
	 * 
	 * @param inARGBPixels		the ARGB pixels to convert
	 * @param outGrayPixels		if not <tt>null</tt>, the output grayscale
	 * 							pixels are stored here; may be same as <tt>inARGBPixels</tt>
	 * @return					the output grayscale pixels
	 */
	public static int[] ARGBPixelsToGrayPixels(int[] inARGBPixels, int[] outGrayPixels) {
		int numPixels = inARGBPixels.length;
		
		// Ensure output buffer is allocated and of the correct size
		if ((outGrayPixels == null) || (outGrayPixels.length != numPixels)) {
			outGrayPixels = new int[numPixels];
		}
		
		// Convert color pixels to gray pixels
		for (int i=0; i<numPixels; i++) {
			outGrayPixels[i] = Pixels.ARGBToGray(inARGBPixels[i]);
		}
		
		return outGrayPixels;
	}
}